@@ -1,5 +1,6 @@
 /* import.c
- * Copyright (C) 1998, 1999, 2000, 2001, 2002 Free Software Foundation, Inc.
+ * Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003
+ *               2004 Free Software Foundation, Inc.
  *
  * This file is part of GnuPG.
  *
@@ -60,7 +61,6 @@ static int import( IOBUF inp, int fast, const char* fname,
                    struct stats_s *stats, unsigned int options );
 static int read_block( IOBUF a, PACKET **pending_pkt, KBNODE *ret_root );
 static void revocation_present(KBNODE keyblock);
-static void remove_bad_stuff (KBNODE keyblock);
 static int import_one( const char *fname, KBNODE keyblock, int fast,
                        struct stats_s *stats, unsigned int options);
 static int import_secret_one( const char *fname, KBNODE keyblock,
@@ -87,13 +87,7 @@ static int merge_keysigs( KBNODE dst, KBNODE src, int *n_sigs,
 int
 parse_import_options(char *str,unsigned int *options)
 {
-  char *tok;
-  int hit=0;
-  struct
-  {
-    char *name;
-    unsigned int bit;
-  } import_opts[]=
+  struct parse_options import_opts[]=
     {
       {"allow-local-sigs",IMPORT_ALLOW_LOCAL_SIGS},
       {"repair-hkp-subkey-bug",IMPORT_REPAIR_PKS_SUBKEY_BUG},
@@ -101,34 +95,7 @@ parse_import_options(char *str,unsigned int *options)
       {NULL,0}
     };
 
-  while((tok=strsep(&str," ,")))
-    {
-      int i,rev=0;
-
-      if(ascii_strncasecmp("no-",tok,3)==0)
-	{
-	  rev=1;
-	  tok+=3;
-	}
-
-      for(i=0;import_opts[i].name;i++)
-	{
-	  if(ascii_strcasecmp(import_opts[i].name,tok)==0)
-	    {
-	      if(rev)
-		*options&=~import_opts[i].bit;
-	      else
-		*options|=import_opts[i].bit;
-	      hit=1;
-	      break;
-	    }
-	}
-
-      if(!hit && !import_opts[i].name)
-	return 0;
-    }
-
-  return hit;
+  return parse_options(str,options,import_opts);
 }
 
 void *
@@ -197,6 +164,8 @@ import_keys( char **fnames, int nnames, int fast,
 	else {
 	    int rc = import( inp, fast, fname, stats, options );
 	    iobuf_close(inp);
+            /* must invalidate that ugly cache to actually close it */
+            iobuf_ioctl (NULL, 2, 0, (char*)fname);
 	    if( rc )
 		log_error("import from `%s' failed: %s\n", fname,
 							   g10_errstr(rc) );
@@ -247,7 +216,6 @@ import( IOBUF inp, int fast, const char* fname,
     }
 
     while( !(rc = read_block( inp, &pending_pkt, &keyblock) )) {
-        remove_bad_stuff (keyblock);
 	if( keyblock->pkt->pkttype == PKT_PUBLIC_KEY )
 	    rc = import_one( fname, keyblock, fast, stats, options );
 	else if( keyblock->pkt->pkttype == PKT_SECRET_KEY ) 
@@ -383,17 +351,17 @@ read_block( IOBUF a, PACKET **pending_pkt, KBNODE *ret_root )
 	/* make a linked list of all packets */
 	switch( pkt->pkttype ) {
 	  case PKT_COMPRESSED:
-	    if( pkt->pkt.compressed->algorithm < 1
-		|| pkt->pkt.compressed->algorithm > 2 ) {
+	    if(check_compress_algo(pkt->pkt.compressed->algorithm))
+	      {
 		rc = G10ERR_COMPR_ALGO;
 		goto ready;
-	    }
-	    {
+	      }
+	    else
+	      {
 		compress_filter_context_t *cfx = m_alloc_clear( sizeof *cfx );
-		cfx->algo = pkt->pkt.compressed->algorithm;
 		pkt->pkt.compressed->buf = NULL;
-		iobuf_push_filter2( a, compress_filter, cfx, 1 );
-	    }
+		push_compress_filter2(a,cfx,pkt->pkt.compressed->algorithm,1);
+	      }
 	    free_packet( pkt );
 	    init_packet(pkt);
 	    break;
@@ -437,22 +405,6 @@ read_block( IOBUF a, PACKET **pending_pkt, KBNODE *ret_root )
     return rc;
 }
 
-
-static void
-remove_bad_stuff (KBNODE keyblock)
-{
-    KBNODE node;
-
-    for (node=keyblock; node; node = node->next ) {
-        if( node->pkt->pkttype == PKT_SIGNATURE ) {
-            /* delete the subpackets we used to use for the
-               verification cache */
-            delete_sig_subpkt (node->pkt->pkt.signature->unhashed,
-                               SIGSUBPKT_PRIV_VERIFY_CACHE);
-        }
-    }
-}
-
 /* Walk through the subkeys on a pk to find if we have the PKS
    disease: multiple subkeys with their binding sigs stripped, and the
    sig for the first subkey placed after the last subkey.  That is,
@@ -544,7 +496,7 @@ print_import_ok (PKT_public_key *pk, PKT_secret_key *sk, unsigned int reason)
 }
 
 
-void
+static void
 print_import_check (PKT_public_key * pk, PKT_user_id * id)
 {
     char * buf;
@@ -630,7 +582,8 @@ import_one( const char *fname, KBNODE keyblock, int fast,
 
     /* It's really PKS corruption, not HKP corruption, but I won't
        change the string in stable. */
-    if((options&IMPORT_REPAIR_PKS_SUBKEY_BUG) && fix_pks_corruption(keyblock))
+    if((options&IMPORT_REPAIR_PKS_SUBKEY_BUG) && fix_pks_corruption(keyblock)
+       && opt.verbose)
       log_info(_("key %08lX: HKP subkey corruption repaired\n"),
 	       (ulong)keyid[1]);
 
@@ -652,11 +605,9 @@ import_one( const char *fname, KBNODE keyblock, int fast,
 	  }
 
     if( !delete_inv_parts( fname, keyblock, keyid, options ) ) {
-	if( !opt.quiet ) {
-	    log_info( _("key %08lX: no valid user IDs\n"),
-							(ulong)keyid[1]);
-	    log_info(_("this may be caused by a missing self-signature\n"));
-	}
+        log_error( _("key %08lX: no valid user IDs\n"), (ulong)keyid[1]);
+	if( !opt.quiet )
+	  log_info(_("this may be caused by a missing self-signature\n"));
 	stats->no_user_id++;
 	return 0;
     }
@@ -686,6 +637,9 @@ import_one( const char *fname, KBNODE keyblock, int fast,
 	}
 	if( opt.verbose > 1 )
 	    log_info (_("writing to `%s'\n"), keydb_get_resource_name (hd) );
+
+	collapse_uids(&keyblock);
+
 	rc = keydb_insert_keyblock (hd, keyblock );
         if (rc)
 	   log_error (_("error writing keyring `%s': %s\n"),
@@ -704,7 +658,7 @@ import_one( const char *fname, KBNODE keyblock, int fast,
 
 	/* we are ready */
 	if( !opt.quiet ) {
-	    char *p=get_user_id_printable (keyid);
+	    char *p=get_user_id_native (keyid);
 	    log_info( _("key %08lX: public key \"%s\" imported\n"),
 		      (ulong)keyid[1],p);
 	    m_free(p);
@@ -780,7 +734,7 @@ import_one( const char *fname, KBNODE keyblock, int fast,
 
 	    /* we are ready */
 	    if( !opt.quiet ) {
-	        char *p=get_user_id_printable(keyid);
+	        char *p=get_user_id_native(keyid);
 		if( n_uids == 1 )
 		    log_info( _("key %08lX: \"%s\" 1 new user ID\n"),
 					     (ulong)keyid[1], p);
@@ -815,7 +769,7 @@ import_one( const char *fname, KBNODE keyblock, int fast,
                   print_import_ok (pk, NULL, 0);
 	
 	    if( !opt.quiet ) {
-	        char *p=get_user_id_printable(keyid);
+	        char *p=get_user_id_native(keyid);
 		log_info( _("key %08lX: \"%s\" not changed\n"),
 			  (ulong)keyid[1],p);
 		m_free(p);
@@ -826,11 +780,14 @@ import_one( const char *fname, KBNODE keyblock, int fast,
     }
 
   leave:
+    if(mod_key)
+      revocation_present(keyblock_orig);
+    else if(new_key)
+      revocation_present(keyblock);
+
     release_kbnode( keyblock_orig );
     free_public_key( pk_orig );
 
-    revocation_present(keyblock);
-
     return rc;
 }
 
@@ -946,8 +903,8 @@ import_revoke_cert( const char *fname, KBNODE node, struct stats_s *stats )
     pk = m_alloc_clear( sizeof *pk );
     rc = get_pubkey( pk, keyid );
     if( rc == G10ERR_NO_PUBKEY ) {
-	log_info( _("key %08lX: no public key - "
-		 "can't apply revocation certificate\n"), (ulong)keyid[1]);
+	log_error( _("key %08lX: no public key - "
+		     "can't apply revocation certificate\n"), (ulong)keyid[1]);
 	rc = 0;
 	goto leave;
     }
@@ -997,12 +954,12 @@ import_revoke_cert( const char *fname, KBNODE node, struct stats_s *stats )
 	if( onode->pkt->pkttype == PKT_USER_ID )
 	    break;
 	else if( onode->pkt->pkttype == PKT_SIGNATURE
-		 && onode->pkt->pkt.signature->sig_class == 0x20
-		 && keyid[0] == onode->pkt->pkt.signature->keyid[0]
-		 && keyid[1] == onode->pkt->pkt.signature->keyid[1] ) {
+		 && !cmp_signatures(node->pkt->pkt.signature,
+				    onode->pkt->pkt.signature))
+	  {
 	    rc = 0;
 	    goto leave; /* yes, we already know about it */
-	}
+	  }
     }
 
 
@@ -1017,7 +974,7 @@ import_revoke_cert( const char *fname, KBNODE node, struct stats_s *stats )
     keydb_release (hd); hd = NULL;
     /* we are ready */
     if( !opt.quiet ) {
-        char *p=get_user_id_printable (keyid);
+        char *p=get_user_id_native (keyid);
 	log_info( _("key %08lX: \"%s\" revocation certificate imported\n"),
 					(ulong)keyid[1],p);
 	m_free(p);
@@ -1091,15 +1048,18 @@ chk_self_sigs( const char *fname, KBNODE keyblock,
 		  rc = check_key_signature( keyblock, n, NULL);
 		  if( rc )
 		    {
-		      char *p=utf8_to_native(unode->pkt->pkt.user_id->name,
+		      if( opt.verbose )
+			{
+			  char *p=utf8_to_native(unode->pkt->pkt.user_id->name,
 				      strlen(unode->pkt->pkt.user_id->name),0);
-		      log_info( rc == G10ERR_PUBKEY_ALGO ?
-				_("key %08lX: unsupported public key "
-				  "algorithm on user id \"%s\"\n"):
-				_("key %08lX: invalid self-signature "
-				  "on user id \"%s\"\n"),
-				(ulong)keyid[1],p);
-		      m_free(p);
+			  log_info( rc == G10ERR_PUBKEY_ALGO ?
+				    _("key %08lX: unsupported public key "
+				      "algorithm on user id \"%s\"\n"):
+				    _("key %08lX: invalid self-signature "
+				      "on user id \"%s\"\n"),
+				    (ulong)keyid[1],p);
+			  m_free(p);
+			}
 		    }
 		  else
 		    unode->flag |= 1; /* mark that signature checked */
@@ -1111,17 +1071,19 @@ chk_self_sigs( const char *fname, KBNODE keyblock,
 		 revocation targets, this may need to be revised. */
 
 		if( !knode ) {
-		    log_info( _("key %08lX: no subkey for key binding\n"),
-					    (ulong)keyid[1]);
+		    if(opt.verbose)
+		      log_info( _("key %08lX: no subkey for key binding\n"),
+				(ulong)keyid[1]);
 		    n->flag |= 4; /* delete this */
 		}
 		else {
 		  rc = check_key_signature( keyblock, n, NULL);
 		  if( rc ) {
-		    log_info(  rc == G10ERR_PUBKEY_ALGO ?
+		    if(opt.verbose)
+		      log_info(rc == G10ERR_PUBKEY_ALGO ?
 			    _("key %08lX: unsupported public key algorithm\n"):
-		            _("key %08lX: invalid subkey binding\n"),
-		            (ulong)keyid[1]);
+			    _("key %08lX: invalid subkey binding\n"),
+			    (ulong)keyid[1]);
 		    n->flag|=4;
 		  }
 		  else {
@@ -1132,8 +1094,9 @@ chk_self_sigs( const char *fname, KBNODE keyblock,
 			bsnode->flag|=4; /* Delete the last binding
 					    sig since this one is
 					    newer */
-			log_info(_("key %08lX: removed multiple subkey "
-				   "binding\n"),(ulong)keyid[1]);
+			if(opt.verbose)
+			  log_info(_("key %08lX: removed multiple subkey "
+				     "binding\n"),(ulong)keyid[1]);
 		      }
 
 		      bsnode=n;
@@ -1152,18 +1115,21 @@ chk_self_sigs( const char *fname, KBNODE keyblock,
                  See the comment in getkey.c:merge_selfsigs_subkey for
                  more */
 		if( !knode ) {
-		    log_info( _("key %08lX: no subkey for key revocation\n"),
-					    (ulong)keyid[1]);
+		    if(opt.verbose)
+		      log_info( _("key %08lX: no subkey for key revocation\n"),
+				(ulong)keyid[1]);
 		    n->flag |= 4; /* delete this */
 		}
 		else {
 		  rc = check_key_signature( keyblock, n, NULL);
 		  if( rc ) {
-		    log_info(  rc == G10ERR_PUBKEY_ALGO ?
+		      if(opt.verbose)
+			log_info(rc == G10ERR_PUBKEY_ALGO ?
 			    _("key %08lX: unsupported public key algorithm\n"):
-		            _("key %08lX: invalid subkey revocation\n"),
-			       (ulong)keyid[1]);
-		    n->flag|=4;
+			    _("key %08lX: invalid subkey revocation\n"),
+			    (ulong)keyid[1]);
+
+		      n->flag|=4;
 		  }
 		  else {
 		    /* It's valid, so is it newer? */
@@ -1172,8 +1138,9 @@ chk_self_sigs( const char *fname, KBNODE keyblock,
 			rsnode->flag|=4; /* Delete the last revocation
 					    sig since this one is
 					    newer */
-			log_info(_("key %08lX: removed multiple subkey "
-				   "revocation\n"),(ulong)keyid[1]);
+			if(opt.verbose)
+			  log_info(_("key %08lX: removed multiple subkey "
+				     "revocation\n"),(ulong)keyid[1]);
 		      }
 
 		      rsnode=n;
@@ -1258,18 +1225,18 @@ delete_inv_parts( const char *fname, KBNODE keyblock,
 	     * to import non-exportable signature when we have the
 	     * the secret key used to create this signature - it
 	     * seems that this makes sense */
-	    log_info( _("key %08lX: non exportable signature "
-				    "(class %02x) - skipped\n"),
-				    (ulong)keyid[1],
-				     node->pkt->pkt.signature->sig_class );
+	    if(opt.verbose)
+	      log_info( _("key %08lX: non exportable signature "
+			  "(class %02x) - skipped\n"),
+			(ulong)keyid[1], node->pkt->pkt.signature->sig_class );
 	    delete_kbnode( node );
 	}
 	else if( node->pkt->pkttype == PKT_SIGNATURE
 		 && node->pkt->pkt.signature->sig_class == 0x20 )  {
 	    if( uid_seen ) {
-		log_error( _("key %08lX: revocation certificate "
-				     "at wrong place - skipped\n"),
-				    (ulong)keyid[1]);
+	        if(opt.verbose)
+		  log_info( _("key %08lX: revocation certificate "
+			      "at wrong place - skipped\n"), (ulong)keyid[1]);
 		delete_kbnode( node );
 	    }
 	    else {
@@ -1284,9 +1251,10 @@ delete_inv_parts( const char *fname, KBNODE keyblock,
 		  int rc = check_key_signature( keyblock, node, NULL);
 		  if( rc )
 		    {
-		      log_error( _("key %08lX: invalid revocation "
-				   "certificate: %s - skipped\n"),
-				 (ulong)keyid[1], g10_errstr(rc));
+		      if(opt.verbose)
+			log_info( _("key %08lX: invalid revocation "
+				    "certificate: %s - skipped\n"),
+				  (ulong)keyid[1], g10_errstr(rc));
 		      delete_kbnode( node );
 		    }
 		}
@@ -1296,17 +1264,18 @@ delete_inv_parts( const char *fname, KBNODE keyblock,
 		 (node->pkt->pkt.signature->sig_class == 0x18 ||
 		  node->pkt->pkt.signature->sig_class == 0x28) &&
 		 !subkey_seen ) {
-	    log_error( _("key %08lX: subkey signature "
-			 "in wrong place - skipped\n"),
-		       (ulong)keyid[1]);
+	    if(opt.verbose)
+	      log_info( _("key %08lX: subkey signature "
+			  "in wrong place - skipped\n"), (ulong)keyid[1]);
 	    delete_kbnode( node );
 	}
 	else if( node->pkt->pkttype == PKT_SIGNATURE
 		 && !IS_CERT(node->pkt->pkt.signature))
 	  {
-	    log_error(_("key %08lX: unexpected signature class (0x%02X) -"
-			" skipped\n"),(ulong)keyid[1],
-		      node->pkt->pkt.signature->sig_class);
+	    if(opt.verbose)
+	      log_info(_("key %08lX: unexpected signature class (0x%02X) -"
+			 " skipped\n"),(ulong)keyid[1],
+		       node->pkt->pkt.signature->sig_class);
 	    delete_kbnode(node);
 	  }
 	else if( (node->flag & 4) ) /* marked for deletion */
@@ -1402,15 +1371,16 @@ collapse_uids( KBNODE *keyblock )
 	kid1 = keyid_from_sk( n->pkt->pkt.secret_key, NULL );
     else
 	kid1 = 0;
-    log_info(_("key %08lX: duplicated user ID detected - merged\n"),
-								 (ulong)kid1);
+    if(!opt.quiet)
+      log_info(_("key %08lX: duplicated user ID detected - merged\n"),
+	       (ulong)kid1);
 
     return 1;
 }
 
 /* Check for a 0x20 revocation from a revocation key that is not
-   present.  This gets called without the benefit of merge_xxxx so you
-   can't rely on pk->revkey and friends. */
+   present.  This may be called without the benefit of merge_xxxx so
+   you can't rely on pk->revkey and friends. */
 static void
 revocation_present(KBNODE keyblock)
 {
@@ -1518,23 +1488,25 @@ merge_blocks( const char *fname, KBNODE keyblock_orig, KBNODE keyblock,
 		    break;
 		else if( onode->pkt->pkttype == PKT_SIGNATURE
 			 && onode->pkt->pkt.signature->sig_class == 0x20
-			 && node->pkt->pkt.signature->keyid[0]
-			    == onode->pkt->pkt.signature->keyid[0]
-			 && node->pkt->pkt.signature->keyid[1]
-			    == onode->pkt->pkt.signature->keyid[1] ) {
+			 && !cmp_signatures(onode->pkt->pkt.signature,
+					    node->pkt->pkt.signature))
+		  {
 		    found = 1;
 		    break;
-		}
+		  }
 	    }
 	    if( !found ) {
-	        char *p=get_user_id_printable (keyid);
 		KBNODE n2 = clone_kbnode(node);
 		insert_kbnode( keyblock_orig, n2, 0 );
 		n2->flag |= 1;
                 ++*n_sigs;
-		log_info(_("key %08lX: \"%s\" revocation certificate added\n"),
-					 (ulong)keyid[1],p);
-		m_free(p);
+		if(!opt.quiet)
+		  {
+		    char *p=get_user_id_native (keyid);
+		    log_info(_("key %08lX: \"%s\" revocation "
+			       "certificate added\n"), (ulong)keyid[1],p);
+		    m_free(p);
+		  }
 	    }
 	}
     }
@@ -1563,8 +1535,9 @@ merge_blocks( const char *fname, KBNODE keyblock_orig, KBNODE keyblock,
 		insert_kbnode( keyblock_orig, n2, 0 );
 		n2->flag |= 1;
                 ++*n_sigs;
-		log_info( _("key %08lX: direct key signature added\n"),
-					 (ulong)keyid[1]);
+		if(!opt.quiet)
+		  log_info( _("key %08lX: direct key signature added\n"),
+			    (ulong)keyid[1]);
 	    }
 	}
     }
@@ -1732,20 +1705,12 @@ merge_sigs( KBNODE dst, KBNODE src, int *n_sigs,
 	    || n->pkt->pkt.signature->sig_class == 0x28 )
 	    continue; /* skip signatures which are only valid on subkeys */
 	found = 0;
-	for(n2=dst->next; n2 && n2->pkt->pkttype != PKT_USER_ID; n2 = n2->next){
-	    if( n2->pkt->pkttype == PKT_SIGNATURE
-		&& n->pkt->pkt.signature->keyid[0]
-		   == n2->pkt->pkt.signature->keyid[0]
-		&& n->pkt->pkt.signature->keyid[1]
-		   == n2->pkt->pkt.signature->keyid[1]
-		&& n->pkt->pkt.signature->timestamp
-		   <= n2->pkt->pkt.signature->timestamp
-		&& n->pkt->pkt.signature->sig_class
-		   == n2->pkt->pkt.signature->sig_class ) {
-		found++;
-		break;
+	for(n2=dst->next; n2 && n2->pkt->pkttype != PKT_USER_ID; n2 = n2->next)
+	  if(!cmp_signatures(n->pkt->pkt.signature,n2->pkt->pkt.signature))
+	    {
+	      found++;
+	      break;
 	    }
-	}
 	if( !found ) {
 	    /* This signature is new or newer, append N to DST.
 	     * We add a clone to the original keyblock, because this
