@@ -19,7 +19,7 @@
  */
 
 #include <config.h>
-#if defined (__MINGW32__) || defined (__CYGWIN32__)
+#if defined (_WIN32) || defined (__CYGWIN32__)
  /* This module is only used in this environment */
 
 #include <stdio.h>
